/**
 * Adapter functionality for canon camera.
 */
const EventEmitter = require('events');
const fs = require('fs');
const fsext = require('fs-extra');
const config = require('config');
const HashMap = require('hashmap');
const canonApi = require('@dimensional/napi-canon-cameras');
const dstCanon = require('../libs/canon.class');
const {CanonCamera, CameraStatus, CameraShot} = require("../libs/canon.class");
const utilities = require("../utils/image-utils");
const cameraUtilities = require('../utils/camera-utils');
const logger = require('../adapters/winston.logger');

const events = new EventEmitter();
let cameraInstance = new dstCanon.CanonCamera();
let cameraShots = new HashMap();
const shotsOutputDirectory = config.get('camera.shots.output');
fsext.ensureDir(shotsOutputDirectory);

const MONITORING_PROPERTIES = ['ISOSpeed', 'Tv', 'Av', 'ExposureCompensation', 'WhiteBalance', 'AFMode', 'DriveMode', 'MeteringMode', 'ImageQuality' ];

/**
 * Specific handler for property changed.
 * @param event
 */
function handlePropertyChangedValue(event){

    if( MONITORING_PROPERTIES.includes(event.property.label) ){
        let propertyVal = cameraInstance.canonInstance.getProperty(canonApi.CameraProperty.ID[event.property.label]);
        let temp = propertyVal.value.label.split(".");
        let propLabel = temp.length > 1 ? temp[1] : temp[0]
        cameraInstance.addProperty(event.property.label, { label: propLabel, value: propertyVal.value.value });
    }

}

/**
 * Events Handlers that handle specific events.
 */
/** Handler for: Camera Property changed events **/
events.on(
    canonApi.CameraBrowser.EventName.PropertyChangeValue,
    (event) => {
        logger.debug('Label: %s, Value:  %s', event.property.label, (event.property.hasOwnProperty('value') ? event.property.value : 'N/A'));
        handlePropertyChangedValue(event);
    },
);

/** Handler for: Camera remove/unplugged events **/
events.on(
    canonApi.CameraBrowser.EventName.CameraRemove,
    (event) => {
        logger.info('Camera removed, Camera information: %s', event.camera.description);
        cameraInstance.productName = 'N/A';
        cameraInstance.firmwareVersion = 'N/A';
        cameraInstance.serialNumber = 'N/A';
        cameraInstance.status = CameraStatus.REMOVED;
        cameraInstance.canonInstance = null;
        cameraInstance.properties = {};
    },
);

/** Handler for: Camera add/plugged events **/
events.on(
    canonApi.CameraBrowser.EventName.CameraAdd,
    (event) => {
        logger.info('New camera added, Camera information: %s', event.camera.description);
        cameraInstance.productName = event.camera.description;
        if( (cameraInstance.canonInstance === undefined) || (cameraInstance.canonInstance === null) ){
            cameraInstance = new dstCanon.CanonCamera(canonApi.cameraBrowser.getCamera());
        }
        cameraInstance.status = CameraStatus.ATTACHED;
        cameraInstance.canonInstance.connect(config.get('camera.keepalive'));
    },
);

/** Handler for: Camera connected events **/
events.on(
    canonApi.CameraBrowser.EventName.CameraConnect,
    (event) => {
        logger.info('Camera connected, Camera information: %s', event.camera.description);
        cameraInstance.productName = event.camera.description;
        cameraInstance.status = CameraStatus.CONNECTED;

        let firmwareVersion = cameraInstance.canonInstance.getProperty(canonApi.CameraProperty.ID.FirmwareVersion);
        cameraInstance.firmwareVersion = (firmwareVersion.available) ? firmwareVersion.value : 'N/A';

        let serialNumber = cameraInstance.canonInstance.getProperty(canonApi.CameraProperty.ID.BodyIDEx);
        cameraInstance.serialNumber = (serialNumber.available) ? serialNumber.value : 'N/A';

        cameraInstance.canonInstance.setProperties(
            {
                [canonApi.CameraProperty.ID.SaveTo]: canonApi.Option.SaveTo.Host,
                [canonApi.CameraProperty.ID.ImageQuality]: canonApi.ImageQuality.ID.LargeJPEGFine,
            }
        )
        cleanUpShotsDirectory();
    },
);

/** Handler for: Camera disconnected events **/
events.on(
    canonApi.CameraBrowser.EventName.CameraDisconnect,
    (event) => {
        logger.info('Camera [%s] disconnected', event.camera.description);
        cameraInstance.status = CameraStatus.DISCONNECTED;
    },
);

/** Handler for: Camera live view started events **/
events.on(
    canonApi.CameraBrowser.EventName.LiveViewStart,
    (event) => {
        logger.info('Camera [%s] live view started', event.camera.description);
        cameraInstance.status = CameraStatus.LIVE_VIEW;
        cleanUpShotsDirectory();
    },
);

/** Handler for: Camera live view stop events **/
events.on(
    canonApi.CameraBrowser.EventName.LiveViewStop,
    (event) => {
        logger.info('Camera [%s] live view stopped', event.camera.description);
        cameraInstance.status = CameraStatus.CONNECTED;
        cleanUpShotsDirectory();
    },
);
/** Handler for: Camera FileChangeEvent changed events **/
events.on(
    canonApi.Camera.EventName.FileChangeEvent,
    (event) => {
        try{
            event.file.downloadToFile(shotsOutputDirectory+'/'+cameraInstance.shot.reference + utilities.getImageExtension(event.file.format.label));
            cameraInstance.shot.imageType = utilities.getImageMime(event.file.format.label);
            cameraShots.set(cameraInstance.shot.reference, cameraInstance.shot);
        }catch ( err ){
            logger.error(err);
        }
    },
);

/**
 * Generic handler for camera events.
 */
canonApi.cameraBrowser.setEventHandler(
    (eventName, ...args) => {
        logger.debug('Event [%s] emitted for camera: %s, Property: %s', eventName, args[0].camera.description, (args[0].hasOwnProperty('property') ? args[0].property.label : 'N/A'));
        try{
            if ( (eventName === canonApi.Camera.EventName.FileCreate) || (eventName === canonApi.Camera.EventName.DownloadRequest) ){
                events.emit(canonApi.Camera.EventName.FileChangeEvent, ...args);
            }else{
                events.emit(eventName, ...args);
            }
        }catch (err){
            logger.error(err);
        }

    },
);

logger.debug("Start watching for camera events");
canonApi.watchCameras();

logger.info("Checking if camera is already attached...");
let tempCamera = canonApi.cameraBrowser.getCamera();
if( tempCamera ){
    logger.info("Camera already attached, initializing..");
    cameraInstance = new dstCanon.CanonCamera(tempCamera);
    cameraInstance.canonInstance.connect(config.get('camera.keepalive'));
}else{
    logger.warn("No cameras found, check that camera is attached correctly. Skipping initialization.");
}

/**
 * Before exist callback.
 */
process.on('beforeExit', (code) => {
    logger.info("Disconnected camera if is still connected.");
    if( cameraInstance.canonInstance ){
        cameraInstance.canonInstance.disconnect();
    }
});

/**
 * Private method to clean up directory with camera shots.
 */
function cleanUpShotsDirectory(){
    logger.info('Clean up shots directory %s', shotsOutputDirectory);
    fsext.emptyDirSync(shotsOutputDirectory);
    cameraShots = new HashMap();
}

/**
 * The value to configure ISO Speed.
 * @param value
 */
function configureIsoSpeed(value){
    cameraInstance.canonInstance.setProperties(
        {
            [canonApi.CameraProperty.ID.ISOSpeed]: new canonApi.ISOSensitivity(cameraUtilities.getIsoSensitivity(value)).value,
        }
    )
}

/**
 * Canon Adapter - Exported functionalities.
 */
module.exports = {

    /**
     * Return camera information
     * @returns {{serialNumber: string, productName: string, firmwareVersion: string, status: CameraStatus}}
     */
    getCamera: function(){
        return {
            "productName": cameraInstance.productName,
            "serialNumber": cameraInstance.serialNumber,
            "firmwareVersion": cameraInstance.firmwareVersion,
            "status": cameraInstance.status,
            "properties": cameraInstance.properties
        };
    },

    /**
     * Begin live view.
     * @returns {Promise<unknown>}
     */
    startLive: function(){
        return new Promise((resolve, reject) => {

            if( (cameraInstance.status === CameraStatus.LIVE_VIEW) || (cameraInstance.status === CameraStatus.LIVE_VIEW_PENDING) ){
                logger.warn("Camera live view already running...");
                resolve();
                return;
            }

            if( cameraInstance.status === CameraStatus.CONNECTED ){
                logger.info("Starting camera live view...");
                cameraInstance.status = CameraStatus.LIVE_VIEW_PENDING;
                cameraInstance.canonInstance.startLiveView();
                resolve();
                return;
            }


            logger.warn("Camera is not connected, please check that camera is attached.");
            reject({status: "CAMERA_NOT_CONNECTED"});
        })
    },

    /**
     * Get the view from the camera.
     * @returns {Promise<unknown>}
     */
    getView: function(){

        return new Promise((resolve, reject) => {

            if( (cameraInstance.status === CameraStatus.LIVE_VIEW) ){
                resolve(utilities.getImage(cameraInstance.canonInstance.getLiveViewImage().getDataURL()));
                return;
            }

            logger.info("Camera live view is not started yet.");
            reject({status: "CAMERA_NOT_IN_LIVE_VIEW"});
        });

    },

    /**
     * Take a shot from camera.
     * @returns {Promise<unknown>}
     */
    takeShot: function(){

        return new Promise((resolve, reject) => {

            if( ( (cameraInstance.status === CameraStatus.LIVE_VIEW) || (cameraInstance.status === CameraStatus.CONNECTED) ) ){
                cameraInstance.shot = new CameraShot(utilities.generateRandomString(), undefined);
                cameraShots.set(cameraInstance.shot.reference, cameraInstance.shot);
                cameraInstance.canonInstance.takePicture();
                resolve({
                    reference: cameraInstance.shot.reference
                });
                return;
            }

            logger.warn("Camera is not connected, please check that camera is attached.");
            reject({status: "CAMERA_NOT_CONNECTED"});
        });

    },

    /**
     *
     * @param reference - The reference to return.
     * @returns {Promise<unknown>}
     */
    getShot: function(reference){

        return new Promise((resolve, reject) => {

            let requestedShot = cameraShots.get(reference);
            if( requestedShot === undefined ){
                logger.error("Image reference %s not found in registry..", reference);
                reject();
                return;
            }

            fs.readFile(shotsOutputDirectory+'/'+reference+utilities.getImageExtensionFromMimeType(requestedShot.imageType),{},function(err, file){
                if( err ){
                    logger.error("Image reference %s not found in filesystem..", reference);
                    reject();
                    return;
                }

                resolve({
                    'type':  requestedShot.imageType,
                    'binary': file
                });

            });

        });
    },

    /**
     * Stop live view.
     * @returns {Promise<unknown>}
     */
    stopLive: function(){
        return new Promise((resolve, reject) => {

            if( (cameraInstance.status === CameraStatus.CONNECTED) || (cameraInstance.status === CameraStatus.LIVE_VIEW_PENDING) ){
                logger.warn("Camera live view stop scheduled...");
                resolve();
                return;
            }

            if( (cameraInstance.status === CameraStatus.LIVE_VIEW) ){
                logger.info("Stopping camera live view...");
                cameraInstance.status = CameraStatus.LIVE_VIEW_PENDING;
                cameraInstance.canonInstance.stopLiveView();
                resolve();
                return;
            }

            logger.warn("Camera is not connected, please check that camera is attached.");
            reject({status: "CAMERA_NOT_CONNECTED"});

        })
    },

    /**
     * Clean up the directory with the camera shots.
     */
    cleanUp: function(){
        cleanUpShotsDirectory();
    },

    /**
     * Configure camera.
     * @param configuration {string} configuration parameter
     * @param value {json object} JSon contains the values for the configuration.
     */
    configure: function(configuration, value){

        return new Promise((resolve, reject) => {

            if( !cameraUtilities.isValidProperty(configuration, value) ){
                logger.error('Invalid configuration %s, Value: %s', configuration, value);
                reject({status: "INVALID_CONFIGURATION_PROPERTY"});
                return;
            }

            if( !cameraUtilities.canApplyProperty(cameraInstance.status) ){
                logger.error("Cannot apply configuration, camera is not connected. %s", configuration);
                reject({status: "CAMERA_NOT_CONNECTED"});
                return;
            }

            try{
                switch ( configuration ){
                    case 'ISOSpeed':
                        configureIsoSpeed(value);
                        resolve();
                        return;
                }
            }catch (err){
                logger.error("An error occurred during property configuration", err);
                reject({status: "NO_COMPATIBLE_MODE_FOR_SETTING_PROPERTY"});
                return;
            }

            logger.error("Not valid configuration or not supported %s", configuration);
            reject({status: "CAMERA_CONFIGURATION_UNKNOWN_ERROR"});

        })

    },

    /**
     * Configure camera with multiple properties
     * @param configuration {string} configuration parameter
     * @param value {json object} JSon contains the values for the configuration.
     */
    batchConfigure: function(configurations){
        return new Promise((resolve, reject) => {
            let properties = cameraUtilities.process(configurations);
            if( properties.errors.length > 0 ) {
                reject({status: "CAMERA_INVALID_CONFIGURATIONS", errors: properties.errors});
                return;
            }

            if( !cameraUtilities.canApplyProperty(cameraInstance.status) ){
                logger.error("Cannot apply configurations, camera is not connected.");
                reject({status: "CAMERA_NOT_CONNECTED"});
                return;
            }

            if( properties.properties !== undefined ){
                try {
                    cameraInstance.canonInstance.setProperties(properties.properties);
                }catch ( err ){
                    logger.error("An error occurred during property configuration", err);
                    reject({status: "NO_COMPATIBLE_MODE_FOR_SETTING_PROPERTY"});
                    return;
                }
            }
            resolve();
        });
    }

}
