/**
 * Canon application related objects.
 */

/**
 * Type of the Camera status.
 * @type
 *  {
 *      {
 *          DISCONNECTED: string,
 *          ATTACHED: string,
 *          LIVE_VIEW: string,
 *          CONNECTED: string,
 *          REMOVED: string
 *      }
 *  }
 */
const CameraStatus = {
    ATTACHED: 'ATTACHED',
    DISCONNECTED: 'DISCONNECTED',
    CONNECTED: 'CONNECTED',
    LIVE_VIEW_PENDING: 'LIVE_VIEW_PENDING',
    LIVE_VIEW: 'LIVE_VIEW',
    REMOVED: 'REMOVED',
}

/**
 * Camera shot re-presentation.
 */
class CameraShot {

    /**
     * Constructor
     * @param reference {String} - Reference for the camera shot.
     * @param imageType {String} - The image type.
     */
    constructor(reference, imageType){
        this.reference = reference;
        this.imageType = imageType;
    }

    /**
     * Getter : Reference of the shot.
     * @returns {String}
     */
    get reference(){
        return this._reference;
    }

    /**
     * Reference setter
     * @param reference {String}
     */
    set reference(reference){
        this._reference = reference
    }

    /**
     * Getter : image type.
     * @returns {String}
     */
    get imageType(){
        return this._imageType;
    }

    /**
     * Image type setter
     * @param imageType {String}
     */
    set imageType(imageType){
        this._imageType = imageType
    }

}

/**
 * Canon Camera High level instance.
 */
class CanonCamera {

    /**
     * Constructor with instance.
     * @param instance
     */
    constructor(canonInstance) {
        this.canonInstance = canonInstance;
        this.status = ((canonInstance) ? CameraStatus.ATTACHED : CameraStatus.REMOVED);
        this.shot = undefined;
        this.productName = 'N/A';
        this.firmwareVersion = 'N/A';
        this.serialNumber = 'N/A';
        this.properties = {}
    }

    /**
     * Return the instance (from Canon API) of camera.
     * @returns {Camera}
     */
    get canonInstance(){
        return this._canonInstance;
    }

    /**
     * Setter method
     * @param {*} Canon API instance
     */
    set canonInstance(canonInstance){
        this._canonInstance = canonInstance;
    }

    /**
     * Getter method
     * @returns {CameraStatus}
     */
    get status(){
        return this._status;
    }

    /**
     * Setter method
     * @param {CameraStatus} status - The status.
     */
    set status(status){
        if( status in CameraStatus ){
            if( (status !== CameraStatus.REMOVED) && ( (this.canonInstance === undefined) || (this.canonInstance === null) ) ){
                throw new Error("Cannot switch to value "+status+", no camera instance found!!!");
            }
            this._status = status;
        }else{
            throw new Error("Value "+status+" is not a valid CameraStatus option.");
        }
    }

    /**
     * Return the image shot reference.
     * @returns {CameraShot}
     */
    get shot(){
        return this._shot;
    }

    /**
     * Setter for shot
     * @param {CameraShot} Camera shot information.
     */
    set shot(shot){
        this._shot = shot;
    }

    /**
     * Return the camera product name.
     * @returns {String}
     */
    get productName(){
        return this._productName;
    }

    /**
     * Setter for product name
     * @param {String} Camera product name.
     */
    set productName(productName){
        this._productName = productName;
    }

    /**
     * Return the camera firmware version.
     * @returns {String}
     */
    get firmwareVersion(){
        return this._firmwareVersion;
    }

    /**
     * Setter for firmware version
     * @param {String} - Camera firmware version.
     */
    set firmwareVersion(firmwareVersion){
        this._firmwareVersion = firmwareVersion;
    }

    /**
     * Return the camera serial number.
     * @returns {String}
     */
    get serialNumber(){
        return this._serialNumber;
    }

    /**
     * Setter for camera serial number
     * @param {String} - Camera serial number.
     */
    set serialNumber(serialNumber){
        this._serialNumber = serialNumber;
    }

    /**
     * Return the camera properties.
     * @returns {Object}
     */
    get properties(){
        return this._properties;
    }

    /**
     * Set Properties.
     * @param {String} - Camera serial number.
     */
    set properties(properties){
        this._properties = properties;
    }

    /**
     * Add property
     */
    addProperty(property, value){
        this._properties[property] = value;
    }

};

/**
 * Exporting Camera info and possible camera statuses.
 * @type {{CanonCamera: CanonCamera, CameraStatus: {DISCONNECTED: string, ATTACHED: string, LIVE_VIEW: string, CONNECTED: string, REMOVED: string}}}
 */
module.exports = {
    CameraStatus: CameraStatus,
    CanonCamera: CanonCamera,
    CameraShot: CameraShot
}
