const express = require('express');
const router = express.Router();
const imageUtils = require('../utils/image-utils');
const canonAdapter = require('../adapters/canon.adapter');
const logger = require('../adapters/winston.logger');

/**
 * Endpoint that will return the status of the camera and its information.
 * Endpoint Full Path: /api/camera
 */
router.get('/', function(req, res, next) {
  res.json(canonAdapter.getCamera());
});

/**
 * Endpoint that will enable/start the live view on camera, Http Method: POST
 * Endpoint Full Path: /api/camera/live
 */
router.post('/live', function(req, res, next) {
    canonAdapter.startLive()
        .then(data => {
            res.json({
                status: 'success',
                liveUrlEndpoint: '/api/camera/live'
            });
        })
        .catch(err => {
            res.json({
                status: 'failed',
                reason: err
            });
        });
});

/**
 * Endpoint that will return live view of the camera, Http Method: GET
 * Endpoint Full Path: /api/camera/live
 * In case of no live view - This method will return no image icon.
 */
router.get('/live', function(req, res, next) {

    canonAdapter.getView()
        .then(image => {
            res.writeHead(200, {
                'Content-Type': image.type,
                'Content-Length': image.binary.length
            });
            res.end(image.binary);
        }).catch(err => {
            imageUtils.noImage()
                .then(data => {
                    res.writeHead(200, {
                        'Content-Type': data.type,
                        'Content-Length': data.binary.length
                    });
                    res.end(data.binary);
                })
                .catch(() => {
                    logger.error("Unable to get live image");
                    res.end();
                });
        });

});

/**
 * Endpoint that will disable/stop the live view on camera, Http Method: DELETE
 * Endpoint Full Path: /api/camera/live
 */
router.delete('/live', function(req, res, next) {
    canonAdapter.stopLive()
        .then(data => {
            res.json({
                status: 'success',
            });
        })
        .catch(err => {
            res.json({
                status: 'failed',
                reason: err
            });
        });
});

/**
 * Endpoint that will take a photo shot from the camera, and will return the URL for the application to pick the file.
 * Endpoint Full Path: /api/camera/shots
 */
router.post('/shots', function(req, res, next) {
    canonAdapter.takeShot()
        .then(data => {
            res.json({
                status: 'success',
                reference: data.reference
            });
        })
        .catch(err => {
            res.json({
                status: 'failed',
                reason: err
            });
        });
});

/**
 * Endpoint that will clean the directory containing the camera shots.
 * Endpoint Full Path: /api/camera/shots
 */
router.delete('/shots', function(req, res, next) {
    canonAdapter.cleanUp();
    res.end();
});

/**
 * Endpoint that will serve the camera shot referece.
 * Endpoint Full Path: /api/camera/shots/:reference
 */
router.get('/shots/:reference', function(req, res, next) {
    canonAdapter.getShot(req.params.reference)
        .then(image => {
            res.writeHead(200, {
                'Content-Type': image.type,
                'Content-Length': image.binary.length
            });
            res.end(image.binary);
        })
        .catch(err => {
            res.status(404);
            res.end();
        });
});

/**
 * Endpoint for configuring camera.
 */
router.post('/configure', function(req, res, next) {
    canonAdapter.batchConfigure(req.body)
        .then(resp => {
            res.json({status: 'success'});
        })
        .catch(err => {
            res.status(422);
            res.json({
                status: err.status,
                reason: err.errors
            });
        });
})

/**
 * Endpoint for configuring camera ISO-Speed.
 */
router.post('/configure/isoSpeed', function(req, res, next) {
    canonAdapter.configure('ISOSpeed',req.body.sensitivity)
        .then(resp => {
            res.json({status: 'success'});
        })
        .catch(err => {
            res.status(422);
            res.json({
                status: 'failed',
                reason: err.status
            });
        })
})

module.exports = router;
