const { CameraStatus } = require('../libs/canon.class');
const {ISOSensitivity, ExposureCompensation, Option, ImageQuality} = require("@dimensional/napi-canon-cameras");
const _ = require("underscore");
const canonApi = require("@dimensional/napi-canon-cameras");
const config = require('config');
/**
 * Contains a status that the camera must be on in order to apply a new configuration.
 * @type {(string)[]}
 */
const CAMERA_STATUS_FOR_APPLY_PROPERTY = [ CameraStatus.CONNECTED, CameraStatus.LIVE_VIEW_PENDING, CameraStatus.LIVE_VIEW ];

/**
 *
 */
const PICTURE_STYLE_MAPPINGS = {
    'Standard': 129
}

/**
 * Values that allowed per property.
 * @type {{isoSpeed: number[]}}
 */
const CONFIGURATION_PROPERTY_ALLOWED = config.get("camera.properties.validations");

/**
 * functions for numeric.
 * @param value
 * @returns {boolean}
 */
function isNumeric(value) {
    return /^-?\d+$/.test(value);
}

/**
 * Validate a given property.
 * @param property
 * @param value
 * @returns {boolean|*}
 */
function validateProperty(property, value){
    if( !CONFIGURATION_PROPERTY_ALLOWED.hasOwnProperty(property) ) return false;
    return CONFIGURATION_PROPERTY_ALLOWED[property].includes(value);
}

/**
 * Generate a compatible property for camera.
 */
function getProperty(property){

    switch ( property.property ){
        case 'ISOSpeed':
            return {
                id: canonApi.CameraProperty.ID.ISOSpeed,
                value: new canonApi.ISOSensitivity(parseInt(Object.keys(ISOSensitivity.Values).find(key => ISOSensitivity.Values[key] === property.value))).value
            };
        case 'Tv':
            return {
                id: canonApi.CameraProperty.ID.Tv,
                value: canonApi.ShutterSpeed.forLabel(property.value).value
            };
        case 'Av':
            return {
                id: canonApi.CameraProperty.ID.Av,
                value: canonApi.Aperture.forLabel(property.value).value
            };
        case 'ExposureCompensation':
            return {
                id: canonApi.CameraProperty.ID.ExposureCompensation,
                value: ExposureCompensation.Values[String(property.value)]
            };
        case 'PictureStyle':
            return {
                id: canonApi.CameraProperty.ID.PictureStyle,
                value: PICTURE_STYLE_MAPPINGS[property.value]
            };
        case 'WhiteBalance':
            return {
                id: canonApi.CameraProperty.ID.WhiteBalance,
                value: Option.WhiteBalance[property.value]
            };
        case 'AFMode':
            return {
                id: canonApi.CameraProperty.ID.AFMode,
                value: Option.AFMode[property.value]
            };
        case 'DriveMode':
            return {
                id: canonApi.CameraProperty.ID.DriveMode,
                value: Option.DriveMode[property.value]
            };
        case 'MeteringMode':
            return {
                id: canonApi.CameraProperty.ID.MeteringMode,
                value: Option.MeteringMode[property.value]
            };
        case 'ImageQuality':
            return {
                id: canonApi.CameraProperty.ID.ImageQuality,
                value: ImageQuality.ID[property.value]
            };
        default:
    }
}

module.exports = {

    /**
     *  Validating current property
     * @param property {String} - The property to check.
     * @param value {any} - The value to check.
     */
    isValidProperty: function(property, value){
        return validateProperty(property, value);
    },

    /**
     * Method for validating if a configuration can be applied to camera.
     * @param status {CameraStatus}
     */
    canApplyProperty: function(status){
        return CAMERA_STATUS_FOR_APPLY_PROPERTY.includes(status);
    },

    /**
     * Get iso sensitivity.
     * @param value {number}
     */
    getIsoSensitivity: function(value){
        if( value === 0 ) return value;
        return parseInt(Object.keys(ISOSensitivity.Values).find(key => ISOSensitivity.Values[key] === value));
    },

    /**
     * Process configurations
     */
    process: function(configurations){
        let validations = [];
        _.keys(configurations).forEach(function(property) {
            let propertyVal = (isNumeric(configurations[property]) ? parseInt(configurations[property]) : configurations[property]);
            if ( validateProperty(property, propertyVal) ){
                validations.push({ property: property, valid: true, value: propertyVal });
            }else{
                validations.push({ property: property, valid: false, allowed: CONFIGURATION_PROPERTY_ALLOWED[property] });
            }
        });
        let nonValidProperties = _.filter(validations, function(prop){ return !prop['valid']; });
        if( nonValidProperties.length > 0 ){
            return {
                properties: undefined,
                errors: [ _.map(nonValidProperties, function(prop){ return { property: prop.property, allowed: prop.allowed }; }) ]
            };
        }


        let properties = {}
        _.each(validations, function(validation){
            let propConfig = getProperty(validation);
            properties[propConfig.id] = propConfig.value;
        });
        return {
            properties: properties,
            errors: []
        }
    }

}
